#!/usr/bin/env python3
# Copyright (C) 2019, 2021 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.

import requests
from html.parser import HTMLParser
import json
import os.path
import re
import sys

USAGE = """\
Usage:
  update-user-agents.py
  update-user-agents.py <html-file>
  update-user-agents.py -h | --help
"""

URL = "https://techblog.willshouse.com/2012/01/03/most-common-user-agents/"
SCRIPT_DIR = os.path.dirname(__file__)
OUT_PATH = os.path.join(SCRIPT_DIR, "..", "librecaptcha", "user_agent_data.py")
NUM_ENTRIES = 30

HEADER = """\
# This file was automatically generated by update_user_agents.py using data
# from <https://techblog.willshouse.com/2012/01/03/most-common-user-agents/>.

# flake8: noqa
USER_AGENTS = \
"""

FOOTER = "\n"


class Parser(HTMLParser):
    def __init__(self):
        self.desc_seen = False
        self.result = None
        super().__init__()

    def handle_data(self, data):
        if self.result is not None:
            return
        if data is None:
            return
        if not self.desc_seen:
            self.desc_seen = bool(re.search(r"\bplain-text\b", data))
            return
        if re.match(r"\s*Mozilla/", data):
            self.result = data


def get_agents(data):
    agents = []
    for agent in data.strip().splitlines()[:NUM_ENTRIES]:
        if len(agents) >= NUM_ENTRIES:
            break
        if re.match(r"\b(iPhone|iPad|Android)\b", agent):
            continue
        agents.append(agent)
    return agents


def write_agents(agents, file):
    print(HEADER, file=file, end="")
    json.dump(agents, file, indent=4)
    print(FOOTER, file=file, end="")


def main():
    if len(sys.argv) == 2 and sys.argv[1] in ["-h", "--help"]:
        print(USAGE, end="")
        sys.exit(0)

    if len(sys.argv) > 2:
        print(USAGE, end="", file=sys.stderr)
        sys.exit(1)

    if len(sys.argv) > 1:
        with open(sys.argv[1], encoding="utf8") as f:
            text = f.read()
    else:
        text = requests.get(URL).text

    parser = Parser()
    parser.feed(text)
    with open(OUT_PATH, "w", encoding="utf8") as f:
        agents = get_agents(parser.result)
        write_agents(agents, f)


if __name__ == "__main__":
    main()



# Copyright (C) 2017, 2019 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.

from .errors import UserError, UserExit
from .librecaptcha import get_token, __version__
from .user_agents import random_user_agent
import os
import re
import sys


def get_cmd():
    if not sys.argv:
        return "librecaptcha"
    if sys.argv[0].startswith("./"):
        return sys.argv[0]
    return os.path.basename(sys.argv[0])


CMD = get_cmd()
USAGE = """\
Usage:
  {0} [options] [--] <api-key> <site-url> [<user-agent>]
  {0} -h | --help | --version

Arguments:
     <api-key>  The reCAPTCHA API key to use. This is usually the value of the
                "data-sitekey" HTML attribute.

    <site-url>  The URL of the site that contains the reCAPTCHA challenge.
                Should start with http:// or https://. Everything after the
                hostname is optional. For example: https://example.com

  <user-agent>  A user-agent string. The client that will use the obtained
                reCAPTCHA token should have this user-agent string. If not
                provided, a random user-agent string will be chosen and shown.

Options:
   -g --gui  Use the GTK 3 GUI (as opposed to the CLI).
    --debug  Show debugging information while running.
  -h --help  Show this help message.
  --version  Show the program version.
""".format(CMD)


def usage(file=sys.stdout):
    print(USAGE, end="", file=file)


def usage_error(exit=True):
    usage(sys.stderr)
    if exit:
        sys.exit(1)


class ParsedArgs:
    def __init__(self):
        self.parse_error = None
        self.api_key = None
        self.site_url = None
        self.user_agent = None
        self.gui = False
        self.debug = False
        self.help = False
        self.version = False


class ArgParser:
    def __init__(self, args):
        self.args = args
        self.index = 0
        self.positional_index = 0
        self.parsed = ParsedArgs()
        self.options_done = False
        self.end_early = False

    @property
    def arg(self):
        try:
            return self.args[self.index]
        except IndexError:
            return None

    @property
    def done(self):
        return self.end_early or self.index >= len(self.args)

    def advance(self):
        self.index += 1

    def error(self, message):
        self.parsed.parse_error = message
        self.end_early = True

    def parse_long_option(self, arg):
        body = arg[len("--"):]
        if body == "debug":
            self.parsed.debug = True
            return
        if body == "help":
            self.parsed.help = True
            self.end_early = True
            return
        if body == "version":
            self.parsed.version = True
            self.end_early = True
            return
        if body == "gui":
            self.parsed.gui = True
            return
        self.error("Unrecognized option: {}".format(arg))

    def parse_short_option_char(self, char):
        if char == "h":
            self.parsed.help = True
            self.end_early = True
            return
        if char == "g":
            self.parsed.gui = True
            return
        self.error("Unrecognized option: -{}".format(char))

    def parse_short_option(self, arg):
        body = arg[len("-"):]
        for char in body:
            self.parse_short_option_char(char)

    def try_parse_option(self):
        arg = self.arg
        if arg == "--":
            self.options_done = True
            return True
        if re.match(r"--[^-]", arg):
            self.parse_long_option(arg)
            return True
        if re.match(r"-[^-]", arg):
            self.parse_short_option(arg)
            return True
        return False

    def parse_positional(self):
        arg = self.arg
        if self.positional_index == 0:
            self.parsed.api_key = arg
            return
        if self.positional_index == 1:
            self.parsed.site_url = arg
            return
        if self.positional_index == 2:
            self.parsed.user_agent = arg
            return
        self.error("Unexpected positional argument: {}".format(arg))

    def parse_single(self):
        if not self.options_done and self.try_parse_option():
            return
        self.parse_positional()
        self.positional_index += 1

    def handle_end(self):
        if self.end_early:
            return
        if self.positional_index < 1:
            self.error("Missing positional argument: <api-key>")
            return
        if self.positional_index < 2:
            self.error("Missing positional argument: <site-url>")
            return

    def parse(self):
        while not self.done:
            self.parse_single()
            self.advance()
        self.handle_end()
        return self.parsed


GOT_TOKEN_MSG = """\
Received token. This token should usually be submitted with the form as the
value of the "g-recaptcha-response" field.
"""


def run(args: ParsedArgs):
    random_ua = False
    user_agent = args.user_agent
    if args.user_agent is None:
        random_ua = True
        user_agent = random_user_agent()
    if args.debug:
        print("User-agent string: {}".format(user_agent), file=sys.stderr)

    uvtoken = get_token(
        args.api_key, args.site_url, user_agent,
        gui=args.gui, debug=args.debug,
    )
    print(GOT_TOKEN_MSG)
    if random_ua:
        print("Note: The following user-agent string was used:")
        print(user_agent)
        print()
    print("Token:")
    print(uvtoken)


UNEXPECTED_ERR_MSG = """\
An unexpected error occurred. The exception traceback is shown below:
"""


def run_or_exit(args: ParsedArgs):
    if args.debug:
        return run(args)
    try:
        return run(args)
    except UserExit:
        sys.exit(2)
    except UserError as e:
        print(e.message, file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        print(file=sys.stderr)
        sys.exit(2)
    except Exception:
        print(UNEXPECTED_ERR_MSG, file=sys.stderr)
        raise


def main():
    args = sys.argv[1:]
    parsed = ArgParser(args).parse()
    error = parsed.parse_error

    if error is not None:
        print(error, file=sys.stderr)
        print("For usage information, run: {} --help".format(CMD),
              file=sys.stderr)
        sys.exit(1)

    if parsed.help:
        usage()
        return

    if parsed.version:
        print(__version__)
        return
    run_or_exit(parsed)


if __name__ == "__main__":
    main()



# Copyright (C) 2017, 2019, 2021 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.

from slimit.parser import Parser
from slimit import ast
import requests

import json
import os
import os.path
import re
import sys

SHOW_WARNINGS = False


def make_parser_raw():
    return Parser()


def make_parser_silent():
    # File descriptor hackiness to silence warnings
    null_fd = os.open(os.devnull, os.O_RDWR)
    old_fd = os.dup(2)
    try:
        os.dup2(null_fd, 2)
        return make_parser_raw()
    finally:
        os.dup2(old_fd, 2)
        os.close(null_fd)
        os.close(old_fd)


make_parser = make_parser_silent
if SHOW_WARNINGS:
    make_parser = make_parser_raw


def load_javascript(url, user_agent):
    print("Downloading <{}>...".format(url), file=sys.stderr)
    r = requests.get(url, headers={
        "User-Agent": user_agent,
    })
    return r.text


def extract_strings(javascript):
    print("Extracting strings...", file=sys.stderr)
    # Hack to work around https://github.com/rspivak/slimit/issues/52
    KEYWORDS = r"(?:catch|delete|return|throw)"
    javascript = re.sub(rf"(\.\s*{KEYWORDS})\b", r"\1_", javascript)
    javascript = re.sub(rf"\b({KEYWORDS})(\s*:)", r"'\1'\2", javascript)
    parsed = make_parser().parse(javascript)
    strings = []

    def add_strings(tree, strings):
        if tree is None:
            return

        if not isinstance(tree, (ast.Node, list, tuple)):
            raise TypeError("Unexpected item: {!r}".format(tree))

        if isinstance(tree, ast.String):
            strings.append(tree.value[1:-1])

        children = tree
        if isinstance(tree, ast.Node):
            children = tree.children()

        for child in children:
            add_strings(child, strings)

    add_strings(parsed, strings)
    return strings


def extract_and_save(url, path, version, rc_version, user_agent):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as f:
        print("{}/{}".format(version, rc_version), file=f)
        js = load_javascript(url, user_agent)
        strings = extract_strings(js)
        strings_json = json.dumps(strings)
        print('Saving strings to "{}"...'.format(path), file=sys.stderr)
        f.write(strings_json)
        return strings



# Copyright (C) 2019 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.


class Frontend:
    def __init__(self, recaptcha):
        self.rc = rc = recaptcha
        rc.on_token = self.__handle_token

    def __handle_token(self, token, **kwargs):
        self.on_token(token)

    def on_token(self, token: str, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def run(self, callback=None):
        if callback:
            self.on_token = callback
        self.rc.run()



# Copyright (C) 2017, 2019 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.

from .errors import UserError
from .extract_strings import extract_and_save

from PIL import Image
import requests

from html.parser import HTMLParser
from threading import Thread
from urllib.parse import urlparse
import base64
import io
import json
import os
import os.path
import re
import sys
import time

BASE_URL = "https://www.google.com/recaptcha/api2/"
API_JS_URL = "https://www.google.com/recaptcha/api.js"
JS_URL_TEMPLATE = """\
https://www.gstatic.com/recaptcha/releases/{}/recaptcha__en.js\
"""

STRINGS_VERSION = "0.1.0"
STRINGS_PATH = os.path.join(
    os.path.expanduser("~"), ".cache", "librecaptcha", "cached-strings",
)

DYNAMIC_SELECT_DELAY = 4.5  # seconds
FIND_GOAL_SEARCH_DISTANCE = 10


def get_full_url(url):
    return BASE_URL.rstrip("/") + "/" + url.lstrip("/")


def get_rc_site_url(url):
    parsed = urlparse(url)
    if not parsed.hostname:
        raise UserError("Error: Site URL has no hostname.")
    if not parsed.scheme:
        raise UserError("Error: Site URL has no scheme.")
    if parsed.scheme not in ["http", "https"]:
        raise UserError(
            "Error: Site URL has invalid scheme: {}".format(parsed.scheme),
        )
    port = parsed.port
    if port is None:
        port = {"http": 80, "https": 443}[parsed.scheme]
    return "{}://{}:{}".format(parsed.scheme, parsed.hostname, port)


def rc_base64(string):
    data = string
    if isinstance(string, str):
        data = string.encode()
    return base64.b64encode(data, b"-_").decode().replace("=", ".")


def load_rc_json(text):
    return json.loads(text.split("\n", 1)[1])


def get_meta(pmeta, probable_index):
    if not isinstance(pmeta, list):
        raise TypeError("pmeta is not a list: {!r}".format(pmeta))

    def matches(meta):
        return meta and isinstance(meta, list)

    if probable_index < len(pmeta):
        meta = pmeta[probable_index]
        if matches(meta):
            return meta

    for child in pmeta:
        if matches(child):
            return child
    raise RuntimeError("Could not find meta; pmeta: {!r}".format(pmeta))


def get_rresp(uvresp):
    if not isinstance(uvresp, list):
        raise TypeError("uvresp is not a list: {!r}".format(uvresp))

    for child in uvresp:
        if child and isinstance(child, list) and child[0] == "rresp":
            return child
    return None


def get_js_strings(user_agent, rc_version):
    def get_json():
        with open(STRINGS_PATH) as f:
            version, text = f.read().split("\n", 1)
            if version != "{}/{}".format(STRINGS_VERSION, rc_version):
                raise OSError("Incorrect version: {}".format(version))
            return json.loads(text)

    try:
        return get_json()
    except (OSError, ValueError, json.JSONDecodeError):
        pass

    result = extract_and_save(
        JS_URL_TEMPLATE.format(rc_version), STRINGS_PATH, STRINGS_VERSION,
        rc_version, user_agent,
    )
    print(file=sys.stderr)
    return result


def get_rc_version(user_agent):
    match = re.search(r"/recaptcha/releases/(.+?)/", requests.get(
        API_JS_URL, headers={
            "User-Agent": user_agent,
        },
    ).text)
    if match is None:
        raise RuntimeError("Could not extract version from api.js.")
    return match.group(1)


def get_image(data):
    return Image.open(io.BytesIO(data))


class Solver:
    def __init__(self, recaptcha):
        self.rc = recaptcha

    def on_solved(response, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError


class HasGrid:
    @property
    def num_rows(self):
        return self.dimensions[0]

    @property
    def num_columns(self):
        return self.dimensions[1]

    @property
    def num_tiles(self):
        return self.num_rows * self.num_columns


class DynamicSolver(Solver, HasGrid):
    def __init__(self, recaptcha, pmeta):
        super().__init__(recaptcha)
        self.selections = []
        meta = get_meta(pmeta, 1)
        self.meta = meta
        self.dimensions = (meta[3], meta[4])
        self.tile_index_map = list(range(self.num_tiles))
        self.last_request_map = [0] * self.num_tiles
        self.latest_index = self.num_tiles - 1

    def on_initial_image(self, image, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def on_tile_image(self, index, image, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def run(self):
        self.rc.show_challenge_goal(self.meta)
        self.first_payload()

    def finish(self, block=True):
        if block:
            time.sleep(self.final_timeout)
        self.on_solved(self.selections)

    @property
    def final_timeout(self):
        return max(self.get_timeout(i) for i in range(self.num_tiles))

    def get_timeout(self, index):
        elapsed = time.monotonic() - self.last_request_map[index]
        duration = max(DYNAMIC_SELECT_DELAY - elapsed, 0)
        return duration

    def first_payload(self):
        image = get_image(self.rc.get("payload", api=False, params={
            "c": self.rc.current_token,
            "k": self.rc.api_key,
        }).content)
        self.on_initial_image(image)

    def select_tile(self, index):
        def target():
            time.sleep(self.get_timeout(index))
            self.on_tile_image(index, image)
        image = self.replace_tile(index)
        Thread(target=target, daemon=True).start()

    def replace_tile(self, index):
        real_index = self.tile_index_map[index]
        self.selections.append(real_index)
        r = self.rc.post("replaceimage", data={
            "c": self.rc.current_token,
            "ds": "[{}]".format(real_index),
        })

        self.last_request_map[index] = time.monotonic()
        data = load_rc_json(r.text)
        self.latest_index += 1
        self.tile_index_map[index] = self.latest_index

        self.rc.current_token = data[1]
        replacement_id = data[2][0]

        image = get_image(self.rc.get("payload", api=False, params={
            "c": self.rc.current_token,
            "k": self.rc.api_key,
            "id": replacement_id,
        }).content)
        return image


class MultiCaptchaSolver(Solver, HasGrid):
    def __init__(self, recaptcha, pmeta):
        super().__init__(recaptcha)
        self.selection_groups = []
        self.dimensions = None
        self.challenge_type = None
        self.previous_token = None
        self.previous_id = None
        self.id = "2"
        self.metas = list(get_meta(pmeta, 5)[0])
        self.next_challenge()

    def on_image(self, image, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def run(self):
        self.first_payload()

    def next_challenge(self):
        meta = self.metas.pop(0)
        self.dimensions = (meta[3], meta[4])
        self.rc.show_challenge_goal(meta)

    def select_indices(self, indices):
        self.selection_groups.append(list(sorted(indices)))
        if self.metas:
            self.replace_image()
            return
        self.on_solved(self.selection_groups)

    def first_payload(self):
        image = get_image(self.rc.get("payload", api=False, params={
            "c": self.rc.current_token,
            "k": self.rc.api_key,
        }).content)
        self.on_image(image)

    def replace_image(self):
        selections = self.selection_groups[-1]
        r = self.rc.post("replaceimage", data={
            "c": self.rc.current_token,
            "ds": json.dumps([selections], separators=",:"),
        })

        data = load_rc_json(r.text)
        self.previous_token = self.rc.current_token
        self.rc.current_token = data[1]

        replacement_id = (data[2] or [None])[0]
        self.previous_id = self.id
        self.id = replacement_id
        self.next_challenge()

        image = get_image(self.rc.get("payload", api=False, params={
            "c": self.previous_token,
            "k": self.rc.api_key,
            "id": self.previous_id,
        }).content)
        self.on_image(image)


class ReCaptcha:
    def __init__(self, api_key, site_url, user_agent, debug=False,
                 make_requests=True):
        self.api_key = api_key
        self.site_url = get_rc_site_url(site_url)
        self._debug = debug
        self.co = rc_base64(self.site_url)

        self.first_token = None
        self.current_token = None
        self.user_agent = user_agent

        self.js_strings = None
        self.rc_version = None
        if make_requests:
            self.rc_version = get_rc_version(self.user_agent)
            self.js_strings = get_js_strings(self.user_agent, self.rc_version)

    def on_goal(goal: str, meta, *, raw: str):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def on_token(token: str, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def on_challenge(type: str, **kwargs):
        """Callback (optional); set this attribute in the parent class."""
        pass

    def on_challenge_dynamic(solver: DynamicSolver, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def on_challenge_multicaptcha(solver: MultiCaptchaSolver, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def on_challenge_blocked(type: str, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def on_challenge_unknown(type: str, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def debug(self, *args, **kwargs):
        if self._debug:
            print(*args, file=sys.stderr, **kwargs)

    def find_challenge_goal(self, id, raw=False):
        start = 0
        matching_strings = []

        def try_find():
            nonlocal start
            index = self.js_strings.index(id, start)
            for i in range(FIND_GOAL_SEARCH_DISTANCE):
                next_str = self.js_strings[index + i + 1]
                if re.search(r"\bselect all\b", next_str, re.I):
                    matching_strings.append((i, index, next_str))
            start = index + FIND_GOAL_SEARCH_DISTANCE + 1

        try:
            while True:
                try_find()
        except (ValueError, IndexError):
            pass

        try:
            goal = min(matching_strings)[2]
        except ValueError:
            return None, None

        raw = goal
        plain = raw.replace("<strong>", "").replace("</strong>", "")
        return raw, plain

    def show_challenge_goal(self, meta):
        raw, goal = self.find_challenge_goal(meta[0])
        self.on_goal(goal, meta, raw=raw)

    def get_headers(self, headers):
        headers = headers or {}
        if "User-Agent" not in headers:
            headers["User-Agent"] = self.user_agent
        return headers

    def get(self, url, *, params=None, api=True, headers=None,
            allow_errors=None, **kwargs):
        params = params or {}
        if api:
            params["k"] = self.api_key
            params["v"] = self.rc_version
        headers = self.get_headers(headers)

        r = requests.get(
            get_full_url(url), params=params, headers=headers,
            **kwargs,
        )
        self.debug("[http] [get] {}".format(r.url))
        if not (allow_errors is True or r.status_code in (allow_errors or {})):
            r.raise_for_status()
        return r

    def post(self, url, *, params=None, data=None, api=True, headers=None,
             allow_errors=None, no_debug_response=False, **kwargs):
        params = params or {}
        data = data or {}
        if api:
            params["k"] = self.api_key
            data["v"] = self.rc_version
        headers = self.get_headers(headers)

        r = requests.post(
            get_full_url(url), params=params, data=data, headers=headers,
            **kwargs,
        )
        self.debug("[http] [post] {}".format(r.url))
        self.debug("[http] [post] [data] {!r}".format(data))
        if not no_debug_response:
            self.debug("[http] [post] [response] {}".format(r.text))
        if not (allow_errors is True or r.status_code in (allow_errors or {})):
            r.raise_for_status()
        return r

    def request_first_token(self):
        class Parser(HTMLParser):
            def __init__(p_self):
                p_self.token = None
                super().__init__()

            def handle_starttag(p_self, tag, attrs):
                attrs = dict(attrs)
                if attrs.get("id") == "recaptcha-token":
                    p_self.token = attrs.get("value")

        text = self.get("anchor", params={"co": self.co}).text
        parser = Parser()
        parser.feed(text)

        if not parser.token:
            raise RuntimeError(
                "Could not get first token. Response:\n{}".format(text),
            )

        self.first_token = parser.token
        self.current_token = self.first_token

    def verify(self, response):
        response_text = json.dumps({"response": response}, separators=",:")
        response_b64 = rc_base64(response_text)

        self.debug("Sending verify request...")
        r = self.post("userverify", data={
            "c": self.current_token,
            "response": response_b64,
        })

        uvresp = load_rc_json(r.text)
        self.debug("Got verify response: {!r}".format(uvresp))
        rresp = get_rresp(uvresp)
        uvresp_token = uvresp[1]
        return (uvresp_token, rresp)

    def get_first_rresp(self):
        self.debug("Getting first rresp...")
        r = self.post("reload", data={"reason": "fi", "c": self.first_token})
        rresp = load_rc_json(r.text)
        self.debug("Got first rresp: {!r}".format(rresp))
        return rresp

    def handle_solved(self, response, **kwargs):
        uvtoken, rresp = self.verify(response)
        if rresp is not None:
            self.solve_challenge(rresp)
            return
        if not uvtoken:
            raise RuntimeError("Got neither uvtoken nor new rresp.")
        self.on_token(uvtoken)

    def solve_challenge(self, rresp):
        challenge_type = rresp[5]
        self.debug("Challenge type: {}".format(challenge_type))
        pmeta = rresp[4]
        self.debug("pmeta: {}".format(pmeta))
        self.current_token = rresp[1]
        self.debug("Current token: {}".format(self.current_token))

        solver_class = {
            "dynamic": DynamicSolver,
            "multicaptcha": MultiCaptchaSolver,
        }.get(challenge_type)

        handler = {
            "dynamic": self.on_challenge_dynamic,
            "multicaptcha": self.on_challenge_multicaptcha,
            "default": self.on_challenge_blocked,
            "doscaptcha": self.on_challenge_blocked,
        }.get(challenge_type)

        self.on_challenge(challenge_type)
        if handler is None:
            self.on_challenge_unknown(challenge_type)
            return
        if solver_class is None:
            handler(challenge_type)
            return
        solver = solver_class(self, pmeta)
        solver.on_solved = self.handle_solved
        handler(solver)

    def run(self):
        self.request_first_token()
        rresp = self.get_first_rresp()
        self.solve_challenge(rresp)



# Copyright (C) 2019 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.


class UserError(Exception):
    """A user-facing exception for an expected error condition (e.g., bad
    user-supplied data). When librecaptcha is run as a program, exceptions of
    this type are shown without a traceback unless --debug is passed.
    """
    def __init__(self, message):
        super().__init__(message)

    @property
    def message(self):
        return self.args[0]


class UserExit(UserError):
    """When librecaptcha is run as a program, throwing this exception causes
    the program to terminate. The exception message is not shown by default.
    """
    def __init__(self, message="Program terminated."):
        super().__init__(message)


class GtkImportError(ImportError):
    pass



# Copyright (C) 2017, 2019 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.

from .errors import UserError
from .frontend import Frontend
from PIL import ImageDraw, ImageFont

from threading import Thread, RLock
from queue import Queue
import io
import json
import os
import subprocess
import sys
import time


def get_font(size):
    typefaces = [
        "FreeSans", "LiberationSans-Regular", "DejaVuSans", "Arial", "arial",
    ]
    for typeface in typefaces:
        try:
            return ImageFont.truetype(typeface, size=size)
        except OSError:
            pass
    return ImageFont.load_default()


FONT_SIZE = 16
FONT = get_font(FONT_SIZE)


def read_indices(prompt, max_index):
    while True:
        print(prompt, end="", flush=True)
        line = input()
        try:
            indices = [int(i) - 1 for i in line.split()]
        except ValueError:
            print("Invalid input.")
            continue
        if all(0 <= i < max_index for i in indices):
            return indices
        print("Numbers out of bounds.")


def draw_lines(image, rows, columns):
    draw = ImageDraw.Draw(image)

    def line(p1, p2):
        draw.line([p1, p2], fill=(255, 255, 255), width=2)

    for i in range(1, rows):
        y = image.height * i // rows - 1
        line((0, y), (image.width, y))

    for i in range(1, columns):
        x = image.width * i // columns - 1
        line((x, 0), (x, image.height))


def draw_indices(image, rows, columns):
    draw = ImageDraw.Draw(image, "RGBA")
    for i in range(rows * columns):
        row = i // columns
        column = i % columns
        corner = (
            image.width * column // columns,
            image.height * (row + 1) // rows,
        )
        text_loc = (
            corner[0] + round(FONT_SIZE / 2),
            corner[1] - round(FONT_SIZE * 1.5),
        )

        text = str(i + 1)
        text_size = FONT.getsize(text)
        draw.rectangle([
            (text_loc[0] - round(FONT_SIZE / 10), text_loc[1]), (
                text_loc[0] + text_size[0] + round(FONT_SIZE / 10),
                text_loc[1] + text_size[1] + round(FONT_SIZE / 10),
            ),
        ], fill=(0, 0, 0, 128))
        draw.text(text_loc, str(i + 1), fill=(255, 255, 255), font=FONT)


def print_temporary(string, file=sys.stdout):
    end = "" if file.isatty() else "\n"
    print(string, file=file, end=end, flush=True)


def clear_temporary(file=sys.stdout):
    if not file.isatty():
        return
    print("\r\x1b[K", file=file, end="", flush=True)


class CliSolver:
    def __init__(self, solver):
        self.solver = solver
        self.__image_procs = []
        self.__has_display = (os.name == "posix")

    def __run_display(self, img_bytes):
        return subprocess.Popen(
            ["display", "-"], stdin=subprocess.PIPE,
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
        )

    def __try_display(self, image):
        if not self.__has_display:
            return False
        img_buffer = io.BytesIO()
        image.save(img_buffer, "png")
        img_bytes = img_buffer.getvalue()

        try:
            proc = self.__run_display(img_bytes)
        except FileNotFoundError:
            self.__has_display = False
            return False
        proc.stdin.write(img_bytes)
        proc.stdin.close()
        self.__image_procs.append(proc)
        return True

    def show_image(self, image):
        if not self.__try_display(image):
            image.show()

    def hide_images(self):
        for proc in self.__image_procs:
            proc.terminate()
        self.__image_procs.clear()

    def run(self):
        self.solver.run()


class CliDynamicSolver(CliSolver):
    def __init__(self, solver):
        super().__init__(solver)
        solver.on_initial_image = self.handle_initial_image
        solver.on_tile_image = self.handle_tile_image
        self.image_open = False
        self.image_queue = Queue()
        self.num_pending = 0
        self.lock = RLock()

    def handle_initial_image(self, image, **kwargs):
        solver = self.solver
        num_rows, num_columns = solver.dimensions
        draw_indices(image, num_rows, num_columns)
        self.show_image(image)

        print("Take a look at the grid of tiles that just appeared. ", end="")
        print("({} rows, {} columns)".format(num_rows, num_columns))
        print("Which tiles should be selected?")
        print("(Top-left is 1; bottom-right is {}.)".format(solver.num_tiles))
        indices = read_indices(
            "Enter numbers separated by spaces: ", solver.num_tiles,
        )
        print()
        self.hide_images()
        self.select_initial(indices)
        self.new_tile_loop()
        solver.finish()

    def new_tile_loop(self):
        while self.num_pending > 0:
            print_temporary("Waiting for next image...")
            index, image = self.image_queue.get()
            clear_temporary()
            self.num_pending -= 1
            self.show_image(image)

            print("Take a look at the image that just appeared.")
            accept = input(
                "Should this image be selected? [y/N] ",
            )[:1].lower() == "y"
            print()

            self.hide_images()
            if accept:
                self.select_tile(index)

    # Called from a non-main thread.
    def handle_tile_image(self, index, image, **kwargs):
        self.image_queue.put((index, image))

    def select_initial(self, indices):
        for i, index in enumerate(indices):
            # Avoid sending initial requests simultaneously.
            self.select_tile(index, 0.25 * i)

    def select_tile_sync(self, index):
        self.num_pending += 1
        self.solver.select_tile(index)

    def select_tile(self, index, delay=0):
        def target():
            delay and time.sleep(delay)
            with self.lock:
                self.select_tile_sync(index)
        Thread(target=target, daemon=True).start()


class CliMultiCaptchaSolver(CliSolver):
    def __init__(self, solver):
        super().__init__(solver)
        solver.on_image = self.handle_image

    def handle_image(self, image, **kwargs):
        solver = self.solver
        num_rows, num_columns = solver.dimensions
        draw_lines(image, num_rows, num_columns)
        draw_indices(image, num_rows, num_columns)
        self.show_image(image)

        print("Take a look at the grid of tiles that just appeared. ", end="")
        print("({} rows, {} columns)".format(num_rows, num_columns))
        print("Which tiles should be selected?")
        print("(Top-left is 1; bottom-right is {}.)".format(solver.num_tiles))
        indices = read_indices(
            "Enter numbers separated by spaces: ", solver.num_tiles,
        )
        print()
        self.hide_images()
        solver.select_indices(indices)


BLOCKED_MSG = """\
ERROR: Received challenge type "{}".

This is usually an indication that reCAPTCHA requests from this network are
being blocked.

Try installing Tor (the full installation, not just the browser bundle) and
running this program over Tor with the "torsocks" command.

Alternatively, try waiting a while before requesting another challenge over
this network.
"""


class Cli(Frontend):
    def __init__(self, recaptcha):
        super().__init__(recaptcha)
        rc = recaptcha
        rc.on_goal = self.handle_goal
        rc.on_challenge = self.handle_challenge
        rc.on_challenge_dynamic = self.challenge_dynamic
        rc.on_challenge_multicaptcha = self.challenge_multicaptcha
        rc.on_challenge_blocked = self.challenge_blocked
        rc.on_challenge_unknown = self.challenge_unknown
        self._first = True

    def handle_goal(self, goal, meta, **kwargs):
        if goal:
            print("CHALLENGE OBJECTIVE: {}".format(goal))
            return
        print("WARNING: Could not determine challenge objective.")
        print("Challenge information: {}".format(json.dumps(meta)))

    def handle_challenge(self, ctype, **kwargs):
        if not self._first:
            print("You must solve another challenge.")
            print()
        self._first = False

    def challenge_dynamic(self, solver, **kwargs):
        CliDynamicSolver(solver).run()

    def challenge_multicaptcha(self, solver, **kwargs):
        CliMultiCaptchaSolver(solver).run()

    def challenge_blocked(self, ctype, **kwargs):
        self.raise_challenge_blocked(ctype)

    def challenge_unknown(self, ctype, **kwargs):
        self.raise_challenge_unknown(ctype)

    @classmethod
    def raise_challenge_blocked(cls, ctype):
        print(BLOCKED_MSG.format(ctype), end="")
        raise UserError(
            "Error: Unsupported challenge type: {}.\n".format(ctype) +
            "Requests are most likely being blocked; see the message above.",
        )

    @classmethod
    def raise_challenge_unknown(cls, ctype):
        raise UserError(
            "Error: Got unsupported challenge type: {}\n".format(ctype) +
            "Please file an issue if this problem persists.",
        )



# Copyright (C) 2017, 2019 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.

from . import cli
from .errors import GtkImportError, UserError
from .recaptcha import ReCaptcha

__version__ = "0.6.4"


def _get_gui():
    try:
        from . import gui
    except GtkImportError as e:
        raise UserError(
            "Error: Could not load the GUI. Is PyGObject installed?\n"
            "Try (re)installing librecaptcha[gtk] with pip.\n"
            "For more details, add the --debug option.",
        ) from e
    return gui


def has_gui():
    try:
        from . import gui  # noqa: F401
    except GtkImportError:
        return False
    return True


def get_token(api_key, site_url, user_agent, *, gui=False, debug=False):
    rc = ReCaptcha(api_key, site_url, user_agent, debug=debug)
    ui = (_get_gui().Gui if gui else cli.Cli)(rc)
    uvtoken = None

    def callback(token):
        nonlocal uvtoken
        uvtoken = token
    ui.run(callback)
    return uvtoken



#!/usr/bin/env python3
# Copyright (C) 2017-2018 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.

from librecaptcha.__main__ import main

if __name__ == "__main__":
    main()



# This file was automatically generated by update_user_agents.py using data
# from <https://techblog.willshouse.com/2012/01/03/most-common-user-agents/>.

# flake8: noqa
USER_AGENTS = [
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.2 Safari/605.1.15",
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:78.0) Gecko/20100101 Firefox/78.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36 Edg/87.0.664.75",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_1_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_1_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.2 Safari/605.1.15",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_1_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.16; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64; rv:78.0) Gecko/20100101 Firefox/78.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36 Edg/87.0.664.66",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36 Edg/88.0.705.50",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.1 Safari/605.1.15",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36"
]



# Copyright (C) 2019 cyclopsian (https://github.com/cyclopsian)
# Copyright (C) 2019 taylor.fish <contact@taylor.fish>
#
# This file is part of librecaptcha.
#
# librecaptcha is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# librecaptcha is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with librecaptcha.  If not, see <http://www.gnu.org/licenses/>.

from .cli import Cli
from .errors import UserExit, GtkImportError
from .frontend import Frontend

from threading import Thread, RLock
import html
import json
import math
import re

try:
    import gi
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk, Gdk, GdkPixbuf, GLib
except ImportError as e:
    raise GtkImportError from e

# Disable GTK warnings
GLib.log_set_writer_func(lambda *args: GLib.LogWriterOutput.HANDLED)


def image_to_gdk_pixbuf(image):
    width, height = image.size
    image_bytes = GLib.Bytes.new(image.tobytes())
    has_alpha = image.mode == "RGBA"
    bpp = 4 if has_alpha else 3
    return GdkPixbuf.Pixbuf.new_from_bytes(
        image_bytes, GdkPixbuf.Colorspace.RGB,
        has_alpha, 8, width, height, width * bpp,
    )


class ChallengeDialogWidget(Gtk.Dialog):
    css_provider = None

    def __init__(self):
        super().__init__()
        self.columns = 0
        self.grid = Gtk.Grid.new()
        self.set_title("CAPTCHA Challenge")
        self.set_icon_name("view-refresh-symbolic")

        self.verify = self.add_button("", Gtk.ResponseType.OK)
        context = self.verify.get_style_context()
        context.add_class("suggested-action")
        self.set_resizable(False)

        self.content = content = self.get_content_area()
        content.set_spacing(6)
        for dir in ["top", "right", "bottom", "left"]:
            getattr(content, "set_margin_" + dir)(6)

        self.header = Gtk.Label.new("")
        self.header.set_xalign(0)
        context = self.header.get_style_context()
        context.add_class("challenge-header")
        content.pack_start(self.header, False, False, 0)
        self.load_css()

    def get_goal(self, **kwargs) -> str:
        """Callback; set this attribute in this parent class."""
        raise NotImplementedError

    def get_note(self, **kwargs) -> str:
        """Callback; set this attribute in this parent class."""
        raise NotImplementedError

    def get_verify_label(self, **kwargs) -> str:
        """Callback; set this attribute in this parent class."""
        raise NotImplementedError

    def make_grid_item(self, index, pixbuf, **kwargs) -> object:
        """Callback; set this attribute in this parent class."""
        raise NotImplementedError

    @property
    def formatted_goal(self):
        raw = str(self.get_goal())
        match = re.fullmatch(r"(.*)<strong>(.*)</strong>(.*)", raw)
        if not match:
            return html.escape(raw)
        groups = match.groups()
        return '{}<span size="xx-large">{}</span>{}'.format(
            *map(html.escape, [
                groups[0] and groups[0] + "\n",
                groups[1],
                groups[2],
            ]),
        )

    @classmethod
    def load_css(cls):
        if cls.css_provider is not None:
            return
        cls.css_provider = Gtk.CssProvider.new()
        cls.css_provider.load_from_data("""
            .challenge-button {
                border-radius: 0;
                border-width: 1px;
                padding: 0;
            }
            .challenge-check, .challenge-header {
                color: @theme_selected_fg_color;
                background-image: linear-gradient(
                    @theme_selected_bg_color, @theme_selected_bg_color);
            }
            .challenge-header {
                padding: 12px;
            }
            .challenge-check {
                border-radius: 50%;
            }
        """.encode("utf8"))
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), cls.css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

    def load_initial(self, image, *, rows, columns):
        self.columns = columns
        pixbuf = image_to_gdk_pixbuf(image)
        stride_x = image.width // columns
        stride_y = image.height // rows
        for i in range(rows * columns):
            column = i % columns
            row = i // columns
            cell_pixbuf = GdkPixbuf.Pixbuf.new(
                GdkPixbuf.Colorspace.RGB, pixbuf.get_has_alpha(),
                8, stride_x, stride_y,
            )
            src_x = stride_x * column
            src_y = stride_y * row
            pixbuf.copy_area(
                src_x, src_y, stride_x, stride_y, cell_pixbuf, 0, 0,
            )
            widget = self.make_grid_item(i, cell_pixbuf)
            context = widget.get_style_context()
            context.add_class("challenge-button")
            self.grid.attach(widget, column, row, 1, 1)
        self.content.pack_start(self.grid, True, True, 0)

    def get_grid_item(self, i):
        column = i % self.columns
        row = i // self.columns
        return self.grid.get_child_at(column, row)

    def replace_grid_item(self, i, new):
        column = i % self.columns
        row = i // self.columns
        old = self.grid.get_child_at(column, row)
        self.grid.remove(old)
        self.grid.attach(new, column, row, 1, 1)
        new.show_all()

    def update(self):
        goal = self.formatted_goal
        note = str(self.get_note())
        verify_label = str(self.get_verify_label())
        markup = goal + "\n"
        self.header.set_markup(markup + note)
        self.verify.set_label(verify_label)

    def run(self):
        self.show_all()
        response = super().run()
        self.destroy()
        while Gtk.events_pending():
            Gtk.main_iteration()
        if response != Gtk.ResponseType.OK:
            raise UserExit


class ChallengeToggleButton(Gtk.ToggleButton):
    def __init__(self, pixbuf):
        super().__init__()
        self.pixbuf = pixbuf
        width = pixbuf.get_width()
        height = pixbuf.get_height()
        self.small_pixbuf = pixbuf.scale_simple(
            width * 0.9, height * 0.9, GdkPixbuf.InterpType.BILINEAR,
        )
        self.set_relief(Gtk.ReliefStyle.NONE)
        fixed = Gtk.Fixed.new()
        self.image = Gtk.Image.new_from_pixbuf(pixbuf)
        self.image.set_size_request(width, height)
        self.check = Gtk.Image.new_from_icon_name(
            "object-select-symbolic", Gtk.IconSize.DND,
        )
        self.check.set_pixel_size(24)
        self.check.set_no_show_all(True)
        context = self.check.get_style_context()
        context.add_class("challenge-check")
        fixed.put(self.image, 0, 0)
        fixed.put(self.check, 0, 0)
        self.add(fixed)
        self.connect("toggled", lambda obj: self.toggle_check())

    def toggle_check(self):
        if self.get_active():
            self.check.show()
            self.image.set_from_pixbuf(self.small_pixbuf)
            return
        self.check.hide()
        self.image.set_from_pixbuf(self.pixbuf)


class BaseCaptchaDialog:
    def __init__(self):
        self.dialog = dialog = ChallengeDialogWidget()
        dialog.get_goal = self.__get_goal

    def __get_goal(self, **kwargs):
        return self.get_goal(**kwargs)

    def get_goal(self, **kwargs) -> str:
        """Callback; set this attribute in this parent class."""
        raise NotImplementedError

    def load_initial(self, image, *, rows, columns):
        self.dialog.load_initial(image, rows=rows, columns=columns)
        self.update()

    def update(self):
        self.dialog.update()

    def run(self):
        self.dialog.run()


class DynamicDialog(BaseCaptchaDialog):
    def __init__(self):
        super().__init__()
        dialog = self.dialog
        dialog.get_note = self.get_note
        dialog.get_verify_label = self.get_verify_label
        dialog.make_grid_item = self.make_grid_item

    def on_clicked(self, index, **kwargs):
        """Callback; set this attribute in the parent class."""
        raise NotImplementedError

    def get_note(self, **kwargs):
        return "Click verify once there are none left"

    def get_verify_label(self, **kwargs):
        return "Ver_ify"

    def make_grid_item(self, index, pixbuf, **kwargs):
        button = Gtk.Button.new()
        image = Gtk.Image.new_from_pixbuf(pixbuf)
        button.add(image)
        button.connect("clicked", lambda obj: self.handle_clicked(obj, index))
        context = button.get_style_context()
        context.add_class("challenge-button")
        return button

    def handle_clicked(self, button, index, **kwargs):
        spinner = Gtk.Spinner.new()
        spinner.set_size_request(32, 32)
        left = (button.get_allocated_width() - 32) / 2
        top = (button.get_allocated_height() - 32) / 2
        spinner.set_margin_top(math.floor(top))
        spinner.set_margin_left(math.floor(left))
        spinner.set_margin_bottom(math.ceil(top))
        spinner.set_margin_right(math.ceil(left))
        self.dialog.replace_grid_item(index, spinner)
        spinner.start()
        self.on_clicked(index)

    def replace_tile_image(self, index, image):
        pixbuf = image_to_gdk_pixbuf(image)
        spinner = self.dialog.get_grid_item(index)
        if pixbuf is None:
            spinner.stop()
            return
        button = self.make_grid_item(index, pixbuf)
        self.dialog.replace_grid_item(index, button)


class MultiCaptchaDialog(BaseCaptchaDialog):
    def __init__(self):
        super().__init__()
        self.selected = set()
        dialog = self.dialog
        dialog.get_note = self.get_note
        dialog.get_verify_label = self.get_verify_label
        dialog.make_grid_item = self.make_grid_item

    def get_note(self, **kwargs):
        note = "If there are none, click skip"
        if self.selected:
            note = '<span alpha="30%">{}</span>'.format(note)
        return note

    def get_verify_label(self, **kwargs):
        if self.selected:
            return "Ver_ify"
        return "Sk_ip"

    def make_grid_item(self, index, pixbuf, **kwargs):
        button = ChallengeToggleButton(pixbuf)
        button.connect(
            "toggled", lambda obj: self.handle_toggled(obj, index),
        )
        return button

    def handle_toggled(self, button, index, **kwargs):
        if button.get_active():
            self.selected.add(index)
        else:
            self.selected.remove(index)
        self.update()

    def run(self):
        super().run()
        return list(self.selected)


class GuiSolver:
    def __init__(self, solver, gui: "Gui"):
        self.solver = solver
        self.gui: "Gui" = gui
        self.dialog = None

    def get_goal(self, **kwargs):
        return self.gui.goal

    def prepare_dialog(self, dialog):
        dialog.get_goal = self.get_goal
        return dialog

    def run(self):
        self.solver.run()


class GuiDynamicSolver(GuiSolver):
    def __init__(self, solver, gui):
        super().__init__(solver, gui)
        solver.on_initial_image = self.handle_initial_image
        solver.on_tile_image = self.handle_tile_image
        self.lock = RLock()

    def handle_initial_image(self, image, **kwargs):
        solver = self.solver
        rows, columns = solver.dimensions
        self.dialog = self.prepare_dialog(DynamicDialog())
        self.dialog.on_clicked = self.handle_gui_tile_clicked
        self.dialog.load_initial(image, rows=rows, columns=columns)
        self.dialog.run()
        print("Submitting solution...")
        solver.finish()

    # Called from a non-UI thread.
    def handle_tile_image(self, index, image, **kwargs):
        def callback():
            self.dialog.replace_tile_image(index, image)
        GLib.timeout_add(0, callback)

    def handle_gui_tile_clicked(self, index):
        def target():
            with self.lock:
                self.solver.select_tile(index)
        Thread(target=target, daemon=True).start()


class GuiMultiCaptchaSolver(GuiSolver):
    def __init__(self, solver, gui):
        super().__init__(solver, gui)
        solver.on_image = self.handle_image

    def handle_image(self, image, **kwargs):
        solver = self.solver
        rows, columns = solver.dimensions
        self.dialog = self.prepare_dialog(MultiCaptchaDialog())
        self.dialog.load_initial(image, rows=rows, columns=columns)
        indices = self.dialog.run()
        print("Submitting solution...")
        solver.select_indices(indices)


class Gui(Frontend):
    def __init__(self, recaptcha):
        super().__init__(recaptcha)
        self.goal = None
        self._first = True
        rc = self.rc
        rc.on_goal = self.handle_goal
        rc.on_challenge = self.handle_challenge
        rc.on_challenge_dynamic = self.challenge_dynamic
        rc.on_challenge_multicaptcha = self.challenge_multicaptcha
        rc.on_challenge_blocked = self.challenge_blocked
        rc.on_challenge_unknown = self.challenge_unknown

    def handle_goal(self, goal, meta, raw, **kwargs):
        if not raw:
            print("WARNING: Could not determine challenge objective.")
            raw = json.dumps(meta)
        self.goal = raw

    def handle_challenge(self, ctype, **kwargs):
        if not self._first:
            print("You must solve another challenge.")
            print()
        self._first = False

    def challenge_dynamic(self, solver, **kwargs):
        GuiDynamicSolver(solver, self).run()

    def challenge_multicaptcha(self, solver, **kwargs):
        GuiMultiCaptchaSolver(solver, self).run()

    def challenge_blocked(self, ctype, **kwargs):
        Cli.raise_challenge_blocked(ctype)

    def challenge_unknown(self, ctype, **kwargs):
        Cli.raise_challenge_unknown(ctype)
